#-----------------------------------------------------------------------------#
# University of Milan: DAPS&CO 2021
# - Social Network Analysis
# Instructor: Rob Fahey <robfahey@aoni.waseda.jp>
#-----------------------------------------------------------------------------#

Sys.setlocale("LC_ALL", 'en_GB.UTF-8')
Sys.setenv(LANG = "en_US.UTF-8")
getwd()
setwd("~/Dropbox/Milan/2021 Social Network Analysis/Class 3 - Week 1 Wednesday/")   
    # Set the above to the folder on your computer where you've saved the files for this class.


# There are many different packages for doing social network analysis in R.
# We're mostly going to be using igraph, so let's install that for now.
# (Other popular options include "sna" and "network".)

install.packages('igraph')


#-----------------------------------------------------------------------------#
# 1. Creating a Simple Graph
#-----------------------------------------------------------------------------#

library(igraph)

# The most simple way to create a graph in R is just to specify the number of nodes it contains and
# then give a list of all the edges between those nodes.

# With the command below, we create a graph with three nodes (n = 3), and three edges. Each edge
# is specified by two values in the vector - the source and destination node.

# We also tell igraph that this graph is undirected - all edges are bidirectional / reciprocal.

my_graph <- graph( n = 3, edges = c(1, 2, 2, 3, 3, 1), directed = FALSE)

# When we look at the resulting graph object, we see some information about the graph and a list of its edges.

show(my_graph)

# We can also make a very simple plot of the graph. Hey look, our first graph visualisation! :)

plot(my_graph)



# Note that if you specify nodes that aren't in the list of edges, they will appear as "orphans", or "isolates".

my_graph <- graph( n = 6, edges = c(1, 2, 2, 3, 3, 1), directed = FALSE)
plot(my_graph)


# We can also specify names for the nodes, instead of just using numbers. 
# Note that when you're using named nodes, you don't need to specify the number of nodes in the graph;
# igraph will figure this out from the list of edges you give it.

my_graph <- graph( c("Tom", "Mary", "Tom", "Harry", "Emma", "Harry", "Mary", "Emma", "Harry", "Jill") )
plot(my_graph)

# We didn't tell R that this graph is undirected, so it has automatically assumed that the edges are 
# unidirectional - they run from the source to the destination (e.g. from Tom to Mary, but not vice versa).


# When we're building a named graph like this, we can specify the "orphan" nodes - ones that are in the graph
# but not connected to any other node - using the "isolates" parameter.

my_graph <- graph( c("Tom", "Mary", "Tom", "Harry", "Emma", "Harry", "Mary", "Emma", "Harry", "Jill", "Tom", "Tom"),
                   isolates = c("Rich", "Kate", "Meg") )
plot(my_graph)

# Note something else interesting about this new version of the graph - in it, Tom is connected to... Tom. 
# Networks allow you to have these "self-connections" - it's not common to need it in social network analysis, but
# the functionality is there if you require it for some reason.



#-----------------------------------------------------------------------------#
# 1.1  Literal Graph Descriptions
#-----------------------------------------------------------------------------#

# igraph also allows us to create graphs using a special kind of formatting syntax, in which you
# basically describe all of the graph's links in the command input.

# You use dashes -- to indicate an undirected edge. 
# +-- and --+ are directed edges, and ++ is a bidirectional edge.
# (Note that you can't have a graph that mixes directed and undirected edges!)

# For example:
my_graph <- graph_from_literal(a--+b, b+--c, c++a)
plot(my_graph)

# You can also specify groups of edges with :

my_graph <- graph_from_literal(a--+b:c:d, b+--c, c++a, d++e:f:g)
plot(my_graph)

# You can see how it's possible to build up relatively complex graphs from quite a small number of
# commands using this specification language.

# This is especially useful if you're building a sample graph to test a theory or illustrate a
# hypothesis - it's perhaps less useful when you're working with real world data.



#-----------------------------------------------------------------------------#
# 2.  Creating an Adjacency Matrix 
#-----------------------------------------------------------------------------#

# You might recall that yesterday, we discussed the Adjacency Matrix - a square table which shows
# the relationships between all of the nodes in a graph.

# This is one of the most fundamental data structures for analysing graphs.

# In some situations, you might have some real-world data that you want to turn into a graph, which
# is most easily represented as an adjacency matrix. Let's see how that would look.

# Let's create a matrix with six rows and columns (i.e. six nodes) and fill it with zeros.
adj_matrix <- matrix(0, 6, 6)  
show(adj_matrix)

# Zero values mean there's no between the nodes - right now all of these nodes are orphans.

# We can directly create edges in this network by putting values in the matrix. Let's connect some
# nodes together.

adj_matrix[1,3] <- 1
adj_matrix[3,2] <- 1
adj_matrix[4,3] <- 1
adj_matrix[5,1] <- 1
adj_matrix[6,2] <- 1
adj_matrix[4,5] <- 1

show(adj_matrix)

# Note that this is a DIRECTED network - the edges only go in one direction and are not mirrored,
# e.g. there is an edge from 1 to 3, but no corresponding edge from 3 to 1.

# If you want to create an UNDIRECTED network in an adjacency matrix, you need to ensure that each
# edge you add is mirrored with a reciprocal edge.


#-----------------------------------------------------------------------------#
# 2.1  Calculating Network Size and Density (and Node Degree)
#-----------------------------------------------------------------------------#


# You may already have noticed that once you have an adjacency matrix, it's incredibly easy to
# calculate the Size and Density of the network.

# The network size is just the number of nodes - i.e. the length of one of the matrix' dimensions.

dim(adj_matrix)[1]

# The network density requires us to calculate two values - the number of edges, and the number of
# possible edges. The number of possible edges is just the size of the matrix (i.e. the number of
# nodes squared), minus the diagonal down the middle (the edges 1-1, 2-2, 3-3 etc. don't count):

dim(adj_matrix)[1]^2 - dim(adj_matrix)[1]

# Meanwhile the number of actual edges is just the sum of the non-zero values in the table:

sum(adj_matrix)

# Put it together and we can calculate the network density in one line:

sum(adj_matrix) / (dim(adj_matrix)[1]^2 - dim(adj_matrix)[1])


# So our little sample network has a network size of 6 and a network density of 0.2.
# This summary information can be really useful when you're working with very large networks and
# need a quick statistical way to see whether they're closely interconnected, or have only a few
# sparse connections holding the nodes together.


# Of course, you can also calculate the DEGREE of each node by looking along the row or
# column of that node and calculating the number of non-zero entries.

# Since this is a directed network, looking at the row for each node could give a very
# different value to looking at the column...

sum(adj_matrix[4,])
sum(adj_matrix[,4])

# These different measures are called "indegree" and "outdegree".
# "Indegree" is the number of directed edges which end at a node - i.e. the number of edges
# which point to that node.
# "Outdegree" is the number of directed edges which start at a node - i.e. the number of
# edges which point away from that node.

# In this case, Node 4 has an "indegree" of 0 and an "outdegree" of 2.

# (When we say "degree" without specifying whether we mean indegree or outdegree, we're usually
# referring to indegree.)


#-----------------------------------------------------------------------------#
# 2.2  Turning an Adjacency Matrix into a Graph
#-----------------------------------------------------------------------------#


# igraph has a function which allows you to load an adjacency matrix like this directly
# into it and convert it into a graph.

my_graph <- graph_from_adjacency_matrix(adj_matrix)
plot(my_graph)

# By looking at the plot, we can see intuitively some of the things that our analysis
# of the graph revealed. For example, we saw that the "indegree" of node 4 was 0, while
# the outdegree was 2; and sure enough, here we can see two arrows leading away from 
# that node, but none incoming.

# It's also possible to force this graph to be treated as undirected - i.e. to pretend
# that the adjacency matrix was symmetrical, even when it wasn't:

my_graph <- graph_from_adjacency_matrix(adj_matrix, mode = "undirected")
plot(my_graph)


#-----------------------------------------------------------------------------#
# 3.  Basic Analysis in igraph
#-----------------------------------------------------------------------------#

# Let's see if we can construct the famous "kite network" graph in R.

kite <- graph_from_literal(Diane--Carol:Fernando:Garth:Ed:Beverly:Andre, 
                           Carol--Andre:Fernando, 
                           Fernando--Andre:Garth:Heather, 
                           Garth--Ed:Beverly:Heather,
                           Beverly--Ed:Andre,
                           Heather--Ike,
                           Ike--Jane)
plot(kite)

# That probably looks like a bit of a mess in your plot, but it's the kite alright - we've
# defined all of the network links correctly.

# igraph will let us examine the nodes and edges of the graph directly with the commands V (for vertex) and E (for edge):

V(kite)

E(kite)

# We can also view the adjacency matrix for this graph:

kite[]

# Note that since this is an undirected graph, the adjacency matrix is symmetrical across the central diagonal.

# Unsurprisingly, we can also find out all of the fundamental statistics about our graph using simple
# igraph commands.

# Network Size (i.e. number of nodes):
vcount(kite)

# Network Density
edge_density(kite)

# Note that the edge_density function assumes you're examining an undirected network. In a DIRECTED network, 
# network density needs to be calculated a little more manually - you can use the below command to do it:

ecount(kite)/(vcount(kite)*(vcount(kite)-1))

# Unsurprisingly, the network density of a directed network is half of the density of an equivalent undirected
# network, because a directed network can have twice as many edges in theory (one indegree edge and one 
# outdegree edge for every pair of nodes in the graph).


# We can also find the diameter of the graph (i.e. the maximum distance between two nodes):
diameter(kite)

# If you want to see more information about the diameter, get_diameter shows an example of a path of that maximum length:
get_diameter(kite)

# Finally, to see the mean distance of the network:
mean_distance(kite)


#-----------------------------------------------------------------------------#
# 3.1  Analysing Nodes' Role in the Network
#-----------------------------------------------------------------------------#

# igraph allows us to calculate the centrality measures that tell us about each node's role within the
# network. We can easily do this all at once (although be warned, in a huge graph these could take
# quite a long time to calculate).
degree(kite)

# If we were in a directed graph, we could specify whether we want to see the indegree or outdegree by
# specifying "in" or "out" as the mode parameter. (Since this is an undirected graph, the results will be the
# same either way).
degree(kite, mode = "in")

# It's possible to include the values we get from these commands in the visual representation of the graph.
# We do this by telling the plot() command that we want the size of the nodes - vertex.size - to be set from the
# results of one of these analysis commands, like so:

plot(kite, vertex.size = degree(kite))

# This might be a bit hard to see, but you can adjust the node sizes just by specifying a multiplication factor:

plot(kite, vertex.size = degree(kite)*5)

# Plotting the graph in this way makes it pretty easy to see which nodes have the highest Degree Centrality.


# We can calculate closeness centrality in the same way. Closeness is the inverse of the average distance
# (geodesic) between a node and every other node on the network.
closeness(kite)

plot(kite, vertex.size = closeness(kite)*300)


# Similarly, we can calculate eigenvector centrality.
# This is a measurement of centrality that's proportional to the sum of the centrality of other nodes
# a given node is connected to.
eigen_centrality(kite)$vector

plot(kite, vertex.size = eigen_centrality(kite)$vector*30)

# (Eigenvector centrality has a bunch of other settings and parameters, but we're only interested in seeing
# the actual results here, so we look at $vector - if you remove that part of the command you can see all
# the other data being returned.)

# And finally, the Betweenness of the nodes - the number of paths (geodesics) that run through each node.

betweenness(kite)
plot(kite, vertex.size = betweenness(kite)*3)


#-----------------------------------------------------------------------------#
# 3.2  Analysing Edges' Role in the Network
#-----------------------------------------------------------------------------#

# Like the analysis of centrality for nodes, there is also a very important measurement of the centrality
# of edges in the network.

# Edge Betweenness calculates how many geodesics (shortest paths) run across a given edge in the network.
# Essentially this tells us how important that edge is to the network structure and to the transmission
# of information over the network.

edge_betweenness(kite)

# Like with the nodes, we can plot the betweenness of the edges - this time by setting the edge.width parameter.

plot(kite, edge.width = edge_betweenness(kite))

# This essentially shows us how much "traffic" in the network flows across each edge.


#-----------------------------------------------------------------------------#
# 4.  Node and Edge Properties
#-----------------------------------------------------------------------------#

# Nodes and Edges can have properties, or attributes, attached to them.

# For example, our Kite network has names attached to each node; those are a kind of attribute.

V(kite)$name


# We can add additional attributes to nodes just by assigning them to a vector, like this:

V(kite)$gender <- c("female", "female", "male", "male", "male", "female", "male", "female", "male", "female")


# Just adding our own attributes like this doesn't change the graph analysis - but we can reference them
# to filter out certain groups or to change how we plot the graph.

# For example, we could use the "gender" attribute we just created to change the colour of the nodes:

plot(kite, vertex.color=c( "pink", "skyblue")[1+(V(kite)$gender=="male")])


# Or we could use it to create and view a subgraph which only includes the women in the network:

plot(delete_vertices(kite, V(kite)$gender=="male"))


# Edges also have properties. There can be different types of edges, and they can have different weights.

# Here, we'll set every edge in the network to be the type "email", and set its weight to 10.

E(kite)$type <- "email"
E(kite)$weight <- 5

# Now we could try adding a new edge to the graph, with some attributes:

kite2 <- kite + edge("Carol", "Ike", type="phone", weight=15)

plot(kite2, edge.width = E(kite2)$weight)


# Incidentally, adding this one edge to the graph pretty dramatically changes its structure...
# Look what happens to Betweenness, for example.

betweenness(kite2)
plot(kite2, vertex.size = betweenness(kite2)*3)

# Wait...  Is that what we expected?
# Why isn't Carol more important in this betweenness graph, when she's got a heavily weighted connection to Ike?

# This is because weights in igraph don't behave the way you might expect.
# A high weight doesn't mean the connection is more IMPORTANT; it means the nodes are more DISTANT.

# So, if we intended this weight to mean "Carol and Ike had 15 phonecalls, compared to the 5 emails other people
# shared", thus implying that Carol and Ike have a stronger connection - we would need to calculate the weight
# as a proportional inverse of some kind.
# (For example, we could set all the other values to be 3 and the Carol--Ike connection to be 1, suggesting they
# are three times closer together than the other connections on the network. In its current form, igraph
# believes them to be three times further apart than the other connections!)



#-----------------------------------------------------------------------------#
# 5.  Importing Network Data
#-----------------------------------------------------------------------------#

# We've now seen a few different ways of building a network inside R and igraph, but what about importing
# one from data that already exists?


# We're going to try importing a dataset which shows all the interactions between characters in the first
# Star Wars movie (Episode IV).

# Movies and novels actually tend to give us really interesting datasets to work with in network analysis;
# often their characters' interactions are very similar to the observed networks of interaction between
# people in the real world, so they're a good source of data to practice on.

# (This particular Star Wars dataset was developed by Evelina Gabasova and Pablo Barbera.)

sw <- read.csv("starwars_iv.csv")
head(sw)

# Okay - we have a pretty standard edgelist here: 
# - Source and Target for each edge; 
# - weight indicating how many scenes the characters are in together.


# igraph has a function called "graph_from_data_frame()" which allows you to load in an edgelist like this
# and creates a graph automatically:

sw_net <- graph_from_data_frame(d = sw, directed = FALSE)

plot(sw_net)


# This isn't the best-looking plot - we'll discuss visualisation a lot more tomorrow. But we could already
# start to perform some straightforward analysis on this network. For example:

degree(sw_net)

# Unsurprisingly, Luke - the central character - has the highest Degree.
# However:

betweenness(sw_net)

# To view this on a graph:
plot(sw_net, vertex.size = betweenness(sw_net))

# Here we see that the droids (C-3PO and R2-D2) have very high betweenness, because they're in some ways the
# narrative device that threads the whole story together - they're present in almost every scene, so they
# link together groups of characters that don't otherwise meet.

# You probably don't even remember who Dodonna was (I had to look him up) - he's the general of the Rebel
# forces, so during the movie, he also interacts with lots of groups (such as the various Rebel pilots)
# who otherwise don't interact with the heroes much. This makes him a highly central character using a
# betweenness measurement.

eigen_centrality(sw_net)$vector
plot(sw_net, vertex.size = eigen_centrality(sw_net)$vector*10)

# Eigenvector Centrality is arguably the most accurate way to establish who the real "central characters"
# of the movie are. Using this, we see the core characters clearly displayed (with Luke, the protagonist,
# as the highest-ranked), and all the other characters are very clearly seen to be side-characters who
# simply orbit around the central group.


#-----------------------------------------------------------------------------#
# In-Class Assignments (If we have time!)
#-----------------------------------------------------------------------------#

# 1) Can you calculate the Network Size and Network Density of the Star Wars network?


# 2) What is the diameter of the network?


# 3) Try to give each character in the network a "type" based on their role in the movie.

# (If you haven't seen Star Wars, try using these catagories: 
#    - Motti, Darth Vader and Tarkin are "imperial", 
#    - Jabba and Greedo are "criminal", 
#    - Beru and Owen are "civilian", 
#    - everyone else on this chart is "rebel")



#-----------------------------------------------------------------------------#
# Today's Assignment
#-----------------------------------------------------------------------------#

# See the assignment PDF file for details.
